import logging
from typing import List, Set, Union

import pandas as pd

from spark_sight.data_references import (
    COL_TASK_DATE_START,
    COL_TASK_DATE_END,
)


def extract_task_info(
    lines_tasks: List[dict],
    stage_ids: Union[List[int], Set[int]],
) -> pd.DataFrame:
    """Extract task information from Spark log lines.
    
    Parameters
    ----------
    lines_tasks : list of dict
        Lines of the Spark log.
    stage_ids : iterable of int
        Stage ids to filter for.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame containing task information.
        It contains the following columns:
        
        * id_task: int
        * id_stage: int
        * date_start__task: start date of the task
        * date_end__task: end date of the task
        * duration_cpu_usage: duration of actual work. Measured in ns
        * duration_cpu_overhead_serde: duration of overhead (de)serialization. Measured in ns
        * duration_cpu_overhead_shuffle: duration of overhead shuffle (reading and writing). Measured in ns

    """
    _log_root = "Extracting task information from Spark event log"

    _perc_log_dict = [0.25, 0.5, 0.75]
    _perc_log_index = 0
    _perc_log_len = len(stage_ids)

    df = None
    
    for stage_id_index, stage_id in enumerate(stage_ids):
        logging.debug(f"Stage {stage_id}")
        
        tasks = extract_events_tasks(lines_tasks, stage_id=stage_id)
        
        for task in tasks:
            _task_info_dict = convert_line_to_metrics(task)
            
            _task_info_df = pd.json_normalize(
                _task_info_dict,
                sep="_",
            )
            
            _task_info_df = _task_info_df.rename(
                columns={
                    "date_start": COL_TASK_DATE_START,
                    "date_end": COL_TASK_DATE_END,
                }
            )
            
            if df is None:
                df = _task_info_df
            else:
                df = pd.concat(
                    [
                        df,
                        _task_info_df,
                    ],
                    ignore_index=True,
                )
        
        _perc = float(stage_id_index) / _perc_log_len
        if (
            _perc_log_index in range(len(_perc_log_dict))
            and _perc > _perc_log_dict[_perc_log_index]
        ):
            logging.info(
                f"{_log_root}: "
                f"{_perc_log_dict[_perc_log_index] * 100:.0f}%"
            )
            _perc_log_index += 1
        
    return df


def extract_events_tasks(
    lines: List[dict],
    stage_id: int,
) -> List[dict]:
    """Extract information related to the stage.
    
    Extracts events `SparkListenerTaskEnd` for the stage
    
    Parameters
    ----------
    lines : list of dict
        Lines of the Spark log.
    stage_id : int
        Stage id.

    Returns
    -------
    list
        List of task events.

    """
    tasks = [
        _ for _ in lines
        if (
            _["Event"] == "SparkListenerTaskEnd"
            and _["Stage ID"] == stage_id
        )
    ]
    
    return tasks


def extract_event_stage(
    lines: List[dict],
    stage_id: int,
) -> dict:
    """Extract information related to the stage.

    Extracts event `SparkListenerStageCompleted`.

    Parameters
    ----------
    lines : list of dict
        Lines of the Spark log.
    stage_id : int
        Stage id.

    Returns
    -------
    dict
        Stage event.

    """
    stage = [
        _ for _ in lines
        if (
            _["Event"] == "SparkListenerStageCompleted"
            and _["Stage Info"]["Stage ID"] == stage_id
        )
    ]
    
    if len(stage) != 1:
        raise ValueError(f"Stage info no exact match {stage}")
    
    return stage[0]


def convert_line_to_metrics(
    task,
) -> dict:
    """Convert Spark log line to nested dict of metrics.
    
    Parameters
    ----------
    task : dict
        Line of the Spark log.

    Returns
    -------
    dict
        Nested dict of task metrics.

    """
    
    # https://spark.apache.org/docs/latest/monitoring.html#executor-task-metrics
    # "Task Metrics" > "Executor CPU Time" does not include time of deserialization,
    # see notes_1.png

    _dict_base = {
        "id": {
            "task": task["Task Info"]["Task ID"],
            "stage": task["Stage ID"],
            "executor": int(task["Task Info"]["Executor ID"]),
        },
        "date": {
            "start": task["Task Info"]["Launch Time"],
            "end": task["Task Info"]["Finish Time"],
        },
        "duration": {
            "cpu": {
                "usage":
                    # Notice this duration is different from the green bar
                    # in the Spark UI. The green bar indicates the duration
                    # of the task being scheduled onto the executor,
                    # not the actual execution on the CPU
                    float(task["Task Metrics"]["Executor CPU Time"]),
                "overhead": {
                    "serde":
                        float(
                            task["Task Metrics"][
                                "Executor Deserialize CPU Time"]
                            # / 1e9
                        )
                        + (
                            float(task["Task Metrics"][
                                "Result Serialization Time"]
                            ) * float(1e6)
                        ),
                    "shuffle": (
                        # "read":
                        float(
                            task["Task Metrics"]["Shuffle Read Metrics"][
                                "Fetch Wait Time"]
                            # / 1e3
                        )
                        # "write":
                        + float(
                            task["Task Metrics"]["Shuffle Write Metrics"][
                                "Shuffle Write Time"]
                            # / 1e3
                        )
                    )
                }
            }
        },
        "memory": {
            "spill": {
                "disk": float(
                    task["Task Metrics"]["Disk Bytes Spilled"]
                ),
                # "memory": float(
                #     task["Task Metrics"]["Memory Bytes Spilled"]
                # ),
            }
        },
    }
    
    _dict_base["date"]["start"] = (
        pd.to_datetime(
            1e6 *
            _dict_base["date"]["start"]
        )
    )
    
    _dict_base["date"]["end"] = (
        pd.to_datetime(
            1e6 *
            _dict_base["date"]["end"]
        )
    )
    
    return _dict_base



from plotly.graph_objs import Figure

import argparse
import json
import logging
import os.path
import sys
import warnings
from json import JSONDecodeError
from pathlib import Path

from plotly.graph_objs import Figure
from plotly.subplots import make_subplots

from spark_sight.create_charts.parsing_spark_history_server import (
    create_chart_efficiency,
    create_chart_stages,
    assign_y_to_stages,
    create_chart_spill,
)
from spark_sight.data_references import (
    COL_ID_EXECUTOR,
)
from spark_sight.data_references import (
    COL_ID_STAGE,
    COL_SUBSTAGE_DATE_START,
    COL_SUBSTAGE_DATE_END,
    COL_SUBSTAGE_DURATION,
    COL_STAGE_DATE_START,
    COL_STAGE_DATE_END,
    COL_SUBSTAGE_DATE_INTERVAL,
)
from spark_sight.log_parse.main import (
    extract_task_info,
)
from spark_sight.log_transform.main import \
    (
    determine_borders_of_stages_asoftasks,
    split_on_borders,
    aggregate_tasks_in_substages,
    create_duration_stage,
)


warnings.simplefilter(action='ignore', category=FutureWarning)


GITHUB_REPO_ADDRESS_HOMEPAGE = "github.com/alfredo-f/spark-sight"
GITHUB_REPO_ADDRESS_ISSUES = f"{GITHUB_REPO_ADDRESS_HOMEPAGE}/issues"


def compose_str_exception_file_invalid_content(
    path_spark_event_log: str,
):
    return (
        f"Content of the provided path \n{path_spark_event_log}\n"
        "does not appear to be a valid Spark event log."
        "\nPlease check the content and retry.\n"
        ""
        f"\nIf you are positive the behavior is not expected,"
        f"\nplease raise an issue at {GITHUB_REPO_ADDRESS_ISSUES}\n"
        ""
    )


root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
root.addHandler(handler)


DEPLOY_MODE_CLUSTER = "cluster_mode"
DEPLOY_MODE_CLIENT = "client_mode"
DEPLOY_MODE_MAP_OUTPUT = {
    DEPLOY_MODE_CLUSTER: "cluster",
    DEPLOY_MODE_CLIENT: "client",
}


def determine_cpus_available(
    cpus: int,
    deploy_mode: str,
) -> int:
    if deploy_mode not in (
        DEPLOY_MODE_CLUSTER,
        DEPLOY_MODE_CLIENT,
    ):
        raise ValueError(
            f"Invalid deploy mode: {deploy_mode}"
        )
    
    cpus_available = cpus

    logging.debug(
        "Subtracting the core reserved for the OS"
    )
    cpus_available -= 1
    
    if deploy_mode == DEPLOY_MODE_CLUSTER:
        logging.debug(
            "Subtracting the core reserved for the driver"
            f" (deploy mode is {deploy_mode})"
        )
        cpus_available -= 1

    return cpus_available


def update_layout(
    fig: Figure,
    id_executor_max: int,
    cmin: float,
    cmax: float,
    df_fig_memory_empty: bool,
):
    _id_executor_max = id_executor_max + 1
    LAYOUT_EFFICIENCY_ROW = 1
    LAYOUT_TIMELINE_SPILL_ROW = 2
    LAYOUT_TIMELINE_STAGES_ROW = 3
    
    _margin = 50
    
    fig.update_layout(
        margin=dict(l=_margin, r=_margin, t=40, b=5),
        title_text="Efficiency: tasks CPU time / available CPU time of cluster",
        title_font_size=20,
        title_pad_l=_margin,
        title_pad_r=_margin,
        title_pad_t=_margin,
        title_pad_b=_margin,
        showlegend=False,
    )

    _id_executor_range = range(1, _id_executor_max)
    
    fig.update_yaxes(
        tickmode='array',
        tickvals=list(_id_executor_range),
        ticktext=[
            f"Executor {_id_executor}"
            for _id_executor in _id_executor_range
        ],
        autorange="reversed",
        row=LAYOUT_TIMELINE_SPILL_ROW,
        col=1,
    )
    
    fig.update_yaxes(
        showticklabels=False,
        autorange="reversed",
        row=LAYOUT_TIMELINE_STAGES_ROW,
        col=1,
    )
    
    fig.update_yaxes(
        tickformat=",.0%",
        range=[0, 1.0],
        row=LAYOUT_EFFICIENCY_ROW,
        col=1,
    )
    
    update_coloraxes_kwargs = dict(
        colorscale=[
            "#e5ecf6",
            "#b300ff",
        ],
        cmin=cmin,
        cmax=cmax,
        showscale=False,
    )
    
    fig.update_coloraxes(
        **update_coloraxes_kwargs
    )
    
    # TODO x in the center of period
    if df_fig_memory_empty:
        annotation_x = (
            min(fig.data[0].x) + (
                (
                    max(fig.data[0].x) - min(fig.data[0].x)
                )
                / 2
            )
        )
        
        annotation_y = (
            id_executor_max / 2
            + 0.5
        )
        
        fig.add_annotation(
            text="No spill, congrats!",
            xref="paper",
            yref="paper",
            x=annotation_x,
            y=annotation_y,
            font=dict(
                size=18,
            ),
            showarrow=False,
            row=LAYOUT_TIMELINE_SPILL_ROW,
            col=1,
        )
    
    
def create_figure(
    df_fig_efficiency,
    df_fig_timeline_stage,
    df_fig_spill,
    cpus_available: int,
    app_info: dict,
):
    _charts_share_timeline = 0.2
    _charts_share_spill = 0.2
    _charts_share_efficiency = 0.6
    assert 1 == sum(
        [
            _charts_share_timeline,
            _charts_share_spill,
            _charts_share_efficiency,
        ]
    )
    
    fig = make_subplots(
        rows=3,
        shared_xaxes=True,
        vertical_spacing=0.02,
        row_width=[
            _charts_share_timeline,
            _charts_share_spill,
            _charts_share_efficiency,
        ],
    )
    
    create_chart_efficiency(
        df_fig_efficiency,
        fig,
        cpus_available=cpus_available,
    )
    
    stages_y = assign_y_to_stages(df_fig_timeline_stage)
    df_fig_timeline_stage.loc[:, "y"] = df_fig_timeline_stage[COL_ID_STAGE].astype(str).replace(
        stages_y
    )
    
    df_fig_timeline_stage.loc[:, "y_labels"] = (
        df_fig_timeline_stage[COL_ID_STAGE].map("{:.0f}".format)
    )

    df_fig_timeline_stage.loc[:, "fig_color"] = (
        "red"
    )
    
    df_fig_spill_empty = df_fig_spill[
        df_fig_spill["memory_spill_disk"] > 0
    ].empty
    
    cmin = 0
    
    if df_fig_spill_empty:
        cmax = 1_000_000
        
    else:
        df_fig_spill = df_fig_spill[
            df_fig_spill["memory_spill_disk"] > 0
        ].copy()

        cmax = df_fig_spill["memory_spill_disk"].max()
        
    id_executor_max = max(
        app_info[COL_ID_EXECUTOR]
    )
    
    # https://plotly.com/python/reference/layout/yaxis/#layout-yaxis-exponentformat
    create_chart_spill(
        df_fig_spill,
        fig,
        col_y=COL_ID_EXECUTOR,
        row=2,
        # Color is discarded because trace is extracted from plotly.express
        px_timeline_color_kwargs=dict(
            color="memory_spill_disk",
        ),
        app_info=app_info,
    )

    create_chart_stages(
        df_fig_timeline_stage,
        fig,
        col_y="y",
        row=3,
    )
    
    update_layout(
        fig,
        id_executor_max=id_executor_max,
        cmin=cmin,
        cmax=cmax,
        df_fig_memory_empty=df_fig_spill_empty,
    )
    
    return fig


def parse_event_log(
    path_spark_event_log: str,
):
    lines_tasks = []
    lines_stages = []

    _log_root = "Parsing Spark event log"
    
    try:
        _file = open(
            path_spark_event_log,
            "r",
        )
        _file_lines = sum(1.0 for _ in _file)
    finally:
        _file.close()
    
    _perc_log_dict = [0.25, 0.5, 0.75]
    _perc_log_index = 0
    
    with open(
        path_spark_event_log,
        "r",
    ) as _file:
        
        for _line_index, _line in enumerate(_file):
            try:
                _line = json.loads(_line)
                
                if _line["Event"] == "SparkListenerStageCompleted":
                    lines_stages.append(_line)
                elif _line["Event"] == "SparkListenerTaskEnd":
                    lines_tasks.append(_line)
            except JSONDecodeError as e:
                logging.debug(f"Invalid line : {_line}")
            
            _perc = float(_line_index) / _file_lines
            if (
                _perc_log_index in range(len(_perc_log_dict))
                and _perc > _perc_log_dict[_perc_log_index]
            ):
                logging.info(
                    f"{_log_root}: "
                    f"{_perc_log_dict[_perc_log_index] * 100:.0f}%"
                )
                _perc_log_index += 1
    
    return (
        lines_tasks,
        lines_stages,
    )


def create_df_fig_efficiency(
    task_info,
    cpus_available,
    borders_of_stages_asoftasks,
):
    metrics = [
        "duration_cpu_usage",
        "duration_cpu_overhead_serde",
        "duration_cpu_overhead_shuffle",
    ]

    task_info_split = split_on_borders(
        task_info,
        borders_all=borders_of_stages_asoftasks,
        metrics=metrics,
    )
    
    task_grouped = aggregate_tasks_in_substages(
        task_info_split,
        borders_all=borders_of_stages_asoftasks,
        metrics=metrics,
        cols_groupby=[COL_SUBSTAGE_DATE_INTERVAL],
    )
    
    task_grouped.loc[:, "duration_agg_cpu_available"] = (
        task_grouped[COL_SUBSTAGE_DURATION]
        * cpus_available
    )
    
    for _metric in metrics:
        task_grouped.loc[:, f"efficiency__{_metric}"] = (
            task_grouped[_metric]
            / task_grouped["duration_agg_cpu_available"]
        )
    
    task_grouped.loc[:, COL_SUBSTAGE_DURATION] = (
        task_grouped[COL_SUBSTAGE_DURATION]
        / float(1e9)
    )
    
    df_fig_efficiency = task_grouped[
        [
            COL_SUBSTAGE_DATE_START,
            COL_SUBSTAGE_DATE_END,
            COL_SUBSTAGE_DURATION,
            COL_ID_STAGE,
        ]
        + list(
            task_grouped.columns[
                task_grouped.columns.str.startswith("efficiency__")
            ]
        )
    ]
    
    return df_fig_efficiency


def create_df_fig_spill(
    task_info,
    borders_of_stages_asoftasks,
):
    metrics = [
        "memory_spill_disk",
    ]
    
    task_info_split = split_on_borders(
        task_info,
        borders_all=borders_of_stages_asoftasks,
        metrics=metrics,
    )
    
    task_grouped = aggregate_tasks_in_substages(
        task_info_split,
        borders_all=borders_of_stages_asoftasks,
        metrics=metrics,
        cols_groupby=[
            COL_ID_EXECUTOR,
            COL_SUBSTAGE_DATE_INTERVAL,
        ]
    )
    
    task_grouped.loc[:, COL_SUBSTAGE_DURATION] = (
        task_grouped[COL_SUBSTAGE_DURATION]
        / float(1e9)
    )
    
    return task_grouped


def _main(
    path_spark_event_log,
    cpus: int,
    deploy_mode: str = DEPLOY_MODE_CLUSTER,
):
    _log_root = "Parsing Spark event log"
    logging.info(f"{_log_root}...")
    
    (
        lines_tasks,
        lines_stages,
    ) = parse_event_log(
        path_spark_event_log,
    )

    logging.info(f"{_log_root}: done\n")
    
    if len(lines_tasks) == 0 or len(lines_stages) == 0:
        logging.critical(
            compose_str_exception_file_invalid_content(
                path_spark_event_log
            )
        )
    
        return
    
    try:
        stage_ids_completed = set(
            _["Stage Info"]["Stage ID"]
            for _ in lines_stages
            if _["Event"] == "SparkListenerStageCompleted"
        )
        
        stage_ids_all = set(
            _["Stage ID"]
            for _ in lines_tasks
        )

        _log_root = "Extracting task information from Spark event log"
        logging.info(f"{_log_root}...")
        
        task_info = extract_task_info(
            lines_tasks=lines_tasks,
            stage_ids=stage_ids_all,
        )
        
        logging.info(f"{_log_root}: done\n")
        
    except Exception:
        logging.critical(
            compose_str_exception_file_invalid_content(path_spark_event_log)
        )
        
        return

    _log_root = "Computing CPU cores available for tasks"
    logging.info(f"{_log_root}...")
    
    cpus_available = determine_cpus_available(
        cpus,
        deploy_mode,
    )
    
    logging.info(f"{_log_root}: done\n")
    
    _log_root = "Determining borders of stages"
    logging.info(f"{_log_root}...")
    
    borders_of_stages_asoftasks = determine_borders_of_stages_asoftasks(
        task_info,
    )
    
    logging.info(f"{_log_root}: done\n")
    
    _log_root = "Creating chart of task efficiency"
    logging.info(f"{_log_root}...")
    
    df_fig_efficiency = create_df_fig_efficiency(
        task_info,
        cpus_available,
        borders_of_stages_asoftasks=borders_of_stages_asoftasks,
    )
    
    logging.info(f"{_log_root}: done\n")
    
    _log_root = "Creating chart of spill"
    logging.info(f"{_log_root}...")
    
    df_fig_spill = create_df_fig_spill(
        task_info,
        borders_of_stages_asoftasks=borders_of_stages_asoftasks,
    )
    
    df_fig_spill.loc[:, COL_ID_EXECUTOR] = (
        df_fig_spill[COL_ID_EXECUTOR].astype(float)
    )

    logging.info(f"{_log_root}: done\n")
    
    _log_root = "Creating chart of stage timeline"
    logging.info(f"{_log_root}...")
    
    df_fig_timeline_stage = (
        create_duration_stage(
            lines_stages,
            stage_ids_completed,
        )
        .rename(
            columns={
                COL_STAGE_DATE_START: COL_SUBSTAGE_DATE_START,
                COL_STAGE_DATE_END: COL_SUBSTAGE_DATE_END,
            }
        )
    )
    
    logging.info(f"{_log_root}: done\n")
    
    app_info = {
        **(
            task_info
            .groupby(lambda _: True)
            .agg(
                {
                    COL_ID_EXECUTOR: "unique",
                }
            )
            .to_dict(orient="records")[0]
        ),
        **(
            df_fig_timeline_stage
            .groupby(lambda _: True)
            .agg(
                {
                    COL_SUBSTAGE_DATE_START: "min",
                    COL_SUBSTAGE_DATE_END: "max",
                }
            )
            .to_dict(orient="records")[0]
        ),
    }
    
    return create_figure(
        df_fig_efficiency,
        df_fig_timeline_stage,
        df_fig_spill,
        cpus_available=cpus_available,
        app_info=app_info,
    )


def main(
    path_spark_event_log: str,
    cpus: int,
    deploy_mode: str = None,
):
    if deploy_mode is None:
        deploy_mode = DEPLOY_MODE_CLUSTER
    
    _path_spark_event_log = Path(path_spark_event_log)
    
    if not os.path.exists(_path_spark_event_log):
        logging.critical(
            f"You provided path \n{path_spark_event_log}\n"
            "but the file does not appear to exist"
            " on your local computer."
            "\nPlease check the spelling and retry.\n"
        )
        
        return
    
    if os.path.getsize(_path_spark_event_log) == 0:
        logging.critical(
            f"The provided path \n{path_spark_event_log}\n"
            "is an empty file."
            "\nPlease provide a valid Spark event log file.\n"
        )
        
        return
    
    fig = _main(
        path_spark_event_log=Path(path_spark_event_log),
        cpus=cpus,
        deploy_mode=deploy_mode,
    )
    
    if fig is not None:
        logging.info("Showing figure...")
        fig.show()
        logging.info("Showing figure: done\n")


def cli_check_positive(value):
    value_int = int(value)
    if value_int <= 0:
        raise argparse.ArgumentTypeError(
            f"Must be greater than zero. Provided {value_int} <= 0"
        )
    return value_int


def main_cli():
    logging.info("")
    
    logging.info(
        r"""
                      _             _       _     _
 ___ _ __   __ _ _ __| | __     ___(_) __ _| |__ | |_
/ __| '_ \ / _` | '__| |/ /____/ __| |/ _` | '_ \| __|
\__ \ |_) | (_| | |  |   <_____\__ \ | (_| | | | | |_
|___/ .__/ \__,_|_|  |_|\_\    |___/_|\__, |_| |_|\__|
    |_|                               |___/
"""
    )
    
    # if not is_latest_version():
    #     logging.info(
    #         "(OMG new version available: pip install --upgrade spark-sight)"
    #         "\n"
    #     )
    
    parser = argparse.ArgumentParser(
        description="Spark performance at a glance."
    )
    parser.add_argument(
        "--path",
        metavar="path",
        help="Local path to the Spark event log",
        dest="path_spark_event_log",
    )
    parser.add_argument(
        "--cpus",
        metavar="cpus",
        help=(
            "Total CPU cores of the cluster"
        ),
        type=cli_check_positive,
    )
    parser.add_argument(
        "--deploy_mode",
        metavar="deploy_mode",
        help=(
            "Deploy mode the Spark application was submitted with"
            ". Defaults to cluster deploy mode"
        ),
        nargs="?",
        const=DEPLOY_MODE_CLUSTER,
        choices=[
            DEPLOY_MODE_CLUSTER,
            DEPLOY_MODE_CLIENT,
        ],
    )
    
    main(
        **vars(parser.parse_args()),
    )



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.graph_objs import Figure

from spark_sight.data_references import (
    COL_ID_STAGE,
    COL_SUBSTAGE_DURATION,
    COL_SUBSTAGE_DATE_START,
    COL_SUBSTAGE_DATE_END,
    COL_ID_EXECUTOR,
)


GANTT_XAXIS_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
GANTT_BARS_DISTANCE = 0.5
GANTT_BARS_HEIGHT = 0.4


def create_chart_efficiency(
    df: pd.DataFrame,
    fig: Figure,
    cpus_available: int,
):
    id_vars = [
        COL_SUBSTAGE_DATE_START,
        COL_SUBSTAGE_DATE_END,
        COL_SUBSTAGE_DURATION,
        COL_ID_STAGE,
    ]
    
    value_vars = list(df.columns[df.columns.str.startswith("efficiency__")])
    
    df_fig = df[
        id_vars + value_vars
    ]
    
    df_fig = (
        df_fig.melt(
            id_vars=id_vars,
            value_vars=value_vars,
        )
        .rename(
            columns={
                "variable": "metric",
            }
        )
    )
    
    _traces = []
    
    metrics_map_text = {
        "efficiency__duration_cpu_usage": "Actual task work",
        "efficiency__duration_cpu_overhead_serde": "Serialization, deserialization",
        "efficiency__duration_cpu_overhead_shuffle": "Shuffle read and write",
    }

    metrics_map_color = {
        "efficiency__duration_cpu_usage": "green",
        "efficiency__duration_cpu_overhead_serde": "violet",
        "efficiency__duration_cpu_overhead_shuffle": "red",
    }
    
    for _metric in df_fig["metric"].unique():
        _df = df_fig[df_fig["metric"] == _metric].copy().reset_index(drop=True)
        _y = _df["value"]
        _len = len(_y)
        
        _hovertext = (
            _df[COL_SUBSTAGE_DATE_START].dt.strftime("%H:%M:%S")
            + pd.Series([" to "] * _len)
            + _df[COL_SUBSTAGE_DATE_END].dt.strftime("%H:%M:%S")
            + pd.Series([" ("] * _len)
            + _df[COL_SUBSTAGE_DURATION].map("{:.0f}".format)
            + pd.Series([" sec)"] * _len)
            + pd.Series(["<br>"] * _len)
            + pd.Series(["<b>"] * _len)
            + pd.Series(_y.apply(lambda _: _ * 100).map("{:.1f}".format))
            + pd.Series(["%</b> ("] * _len)
            + _df["metric"].map(metrics_map_text).astype(str)
            + pd.Series([")"] * _len)
            + pd.Series(["<br>"] * _len)
            + _df[COL_ID_STAGE].map(len).map(lambda _: "Stage" + ("s" if _ > 1 else "") + ": <b>")
            + pd.Series([", ".join([str(__) for __ in _]) for _ in _df[COL_ID_STAGE].values])
            + pd.Series(["</b>"] * _len)
        )

        _hovertemplate = "<br>".join(
            [
                "%{hovertext}",
                "<extra></extra>",
            ]
        )
        
        _traces.append(
            go.Bar(
                x=_df[COL_SUBSTAGE_DATE_START],
                y=_y,
                width=_df[COL_SUBSTAGE_DURATION].apply(lambda _: float(_)) * 1e3,
                name=_metric,
                offset=0,
                hovertext=_hovertext,
                hovertemplate=_hovertemplate,
                marker_color=metrics_map_color[_metric],
            ),
        )
        
    fig.add_traces(
        _traces,
        rows=1,
        cols=1,
    )
    
    fig.update_layout(
        barmode="stack",
    )

    fig.update_yaxes(
        title_text=f"CPU cores available for tasks: {cpus_available}",
        title_font_size=18,
        row=1,
        col=1,
    )


def assign_y_to_stages(
    _df_stages: pd.DataFrame,
):
    _df_stages = _df_stages.sort_values(COL_SUBSTAGE_DATE_START)
    y = dict()
    
    id_stage_to_see = sorted(list(set(_df_stages[COL_ID_STAGE])))
    id_stage_seen = set()
    
    while len(id_stage_to_see) > 0:
        id_stage = id_stage_to_see[0]

        id_stage_to_see = set(id_stage_to_see)
        id_stage_to_see.remove(id_stage)
        id_stage_seen.add(id_stage)
        
        stage_row = _df_stages[_df_stages[COL_ID_STAGE] == id_stage]
        
        if len(stage_row) != 1:
            raise ValueError(
                f"Stage row not found or found multiple: {stage_row}"
            )
        
        _start, _end = (
            stage_row[COL_SUBSTAGE_DATE_START].values[0],
            stage_row[COL_SUBSTAGE_DATE_END].values[0],
        )
        
        started_during_me = (
            _df_stages[
                (~_df_stages[COL_ID_STAGE].isin(id_stage_seen))
                & (
                    _start <= _df_stages[COL_SUBSTAGE_DATE_START]
                )
                & (
                    _df_stages[COL_SUBSTAGE_DATE_START] <= _end
                )
            ]
        )
        
        _concurrent_stages = [id_stage]
        
        while not started_during_me.empty:
            _concurrent_stages.extend(
                list(
                    started_during_me[COL_ID_STAGE].unique()
                )
            )

            _id_stage_seen = list(started_during_me[COL_ID_STAGE].values)
            
            for _seen in _id_stage_seen:
                id_stage_to_see.remove(_seen)
                id_stage_seen.add(_seen)

            _start, _end = (
                started_during_me[COL_SUBSTAGE_DATE_START].max(),
                started_during_me[COL_SUBSTAGE_DATE_END].max(),
            )

            started_during_me = (
                _df_stages[
                    (~_df_stages[COL_ID_STAGE].isin(id_stage_seen))
                    & (
                        _start <= _df_stages[COL_SUBSTAGE_DATE_START]
                    )
                    & (
                        _df_stages[COL_SUBSTAGE_DATE_START] <= _end
                    )
                ]
            )

        _concurrent_stages = sorted(_concurrent_stages)
        
        y = dict(
            **y,
            **{
                str(_stage): GANTT_BARS_DISTANCE * _stage_index
                for _stage_index, _stage in enumerate(_concurrent_stages)
            },
        )
        
        id_stage_to_see = sorted(list(id_stage_to_see))

    return y


def format_bytes(
    size: float,
) -> str:
    format_float_decimals = 1
    
    power = 2**10
    n = 0
    
    power_labels = {
        0: '',
        1: 'K',
        2: 'M',
        3: 'G',
        4: 'T',
    }
    
    while size > power:
        size /= power
        n += 1
    
    n = min(
        n,
        len(power_labels) - 1,
    )
    
    return (
        (
            (
                r"{:." + str(format_float_decimals) + r"f}"
            )
            .format(size)
        )
        + " "
        + power_labels[n] + 'B'
    )


def _create_trace_spill(
    df: pd.DataFrame,
    col_y: str,
    px_timeline_color_kwargs: dict = None,
    opacity: float = 1,
    is_trace_hidden: bool = False,
):
    df = df.reset_index(drop=True)
    
    # Plotly can"t handle ns
    # plotly.express._core == 5.7.0, line 1681
    # args["data_frame"][args["x_end"]] = (x_end - x_start).astype("timedelta64[ms]")
    df.loc[:, COL_SUBSTAGE_DATE_START] = (
        df[COL_SUBSTAGE_DATE_START].dt.strftime(
            GANTT_XAXIS_DATETIME_FORMAT
        )
    )
    
    df.loc[:, COL_SUBSTAGE_DATE_END] = (
        df[COL_SUBSTAGE_DATE_END].dt.strftime(GANTT_XAXIS_DATETIME_FORMAT)
    )
    
    if not is_trace_hidden:
        text_col = "memory_spill_disk_format"
        
        df.loc[:, "memory_spill_disk_format"] = (
            df["memory_spill_disk"].apply(format_bytes)
        )
        
    else:
        text_col = "y_labels"
    
    _fig = px.timeline(
        data_frame=df,
        x_start=COL_SUBSTAGE_DATE_START,
        x_end=COL_SUBSTAGE_DATE_END,
        y=col_y,
        text=text_col,
        **px_timeline_color_kwargs,
    )
    
    _len = len(df)
    _start = pd.to_datetime(
        df[COL_SUBSTAGE_DATE_START],
        format=GANTT_XAXIS_DATETIME_FORMAT
    )
    _end = pd.to_datetime(
        df[COL_SUBSTAGE_DATE_END],
        format=GANTT_XAXIS_DATETIME_FORMAT
    )
    _duration = (_end - _start).astype("timedelta64[s]")
    
    _fig.update_traces(
        opacity=opacity,
        marker_line=dict(
            width=1,
            color="white",
        ),
        textposition="inside",
        textangle=0,
        insidetextanchor="middle",
    )
    
    if not is_trace_hidden:
        
        _hovertext_date_range = (
            _start.dt.strftime("%H:%M:%S")
            + pd.Series([" to "] * _len)
            + _end.dt.strftime("%H:%M:%S")
            + pd.Series([" ("] * _len)
            + _duration.map("{:.0f}".format)
            + pd.Series([" sec)"] * _len)
        )
        
        _hovertext_stage_all = (
            df[COL_ID_STAGE].map(len).map(
                lambda _: "Stage" + ("s" if _ > 1 else "") + ": <b>"
            )
            + pd.Series(
            [", ".join([str(__) for __ in _]) for _ in
                df[COL_ID_STAGE].values]
        )
            + pd.Series(["</b>"] * _len)
        )
        
        _hovertext = (
            _hovertext_date_range
            + pd.Series(["<br>"] * _len)
            + pd.Series(["Executor ID: <b>"] * _len)
            + df[COL_ID_EXECUTOR].astype(int).astype(str)
            + pd.Series(["</b>"] * _len)
            + pd.Series(["<br>"] * _len)
            + pd.Series(["Total spilled: <b>"] * _len)
            + df["memory_spill_disk_format"]
            + pd.Series(["</b>"] * _len)
            + pd.Series(["<br>"] * _len)
            + _hovertext_stage_all
        )
    
        _fig.update_traces(
            hovertext=_hovertext,
            hovertemplate=(
                "%{hovertext}"
                "<extra></extra>"
            ),
        )

    _trace = list(
        _fig.select_traces()
    )
    
    assert len(_trace) == 1
    return _trace[0]
    

def create_chart_spill(
    df: pd.DataFrame,
    fig,
    col_y: str,
    row: int,
    px_timeline_color_kwargs: dict = None,
    app_info: dict = None,
):
    px_timeline_color_kwargs = px_timeline_color_kwargs or {}
    
    df = df[
        df["memory_spill_disk"] > 0
    ].copy().reset_index(drop=True)
    
    if not df.empty:
        trace_not_empty = _create_trace_spill(
            df,
            col_y,
            px_timeline_color_kwargs,
        )

        fig.add_trace(
            trace_not_empty,
            row=row,
            col=1,
        )
    
    executors_all = set(app_info[COL_ID_EXECUTOR])
    
    df_empty = (
        pd.DataFrame(
            [
                {
                    COL_ID_EXECUTOR: _id_executor,
                    COL_SUBSTAGE_DATE_START: app_info[COL_SUBSTAGE_DATE_START],
                    COL_SUBSTAGE_DATE_END: app_info[COL_SUBSTAGE_DATE_START],
                }
                for _id_executor in executors_all
            ]
            + [
                {
                    COL_ID_EXECUTOR: _id_executor,
                    COL_SUBSTAGE_DATE_START: (
                        app_info[COL_SUBSTAGE_DATE_END]
                    ),
                    COL_SUBSTAGE_DATE_END: (
                        app_info[COL_SUBSTAGE_DATE_END]
                    ),
                }
                for _id_executor in executors_all
            ]
        )
    )
    
    df_empty.loc[:, "y_labels"] = ""
    
    trace_empty = _create_trace_spill(
        df_empty,
        col_y,
        px_timeline_color_kwargs={},
        opacity=0,
        is_trace_hidden=True,
    )
    
    fig.add_trace(
        trace_empty,
        row=row,
        col=1,
    )

    fig.update_xaxes(type="date")


def create_chart_stages(
    df: pd.DataFrame,
    fig,
    col_y: str,
    row: int,
    px_timeline_color_kwargs: dict = None,
):
    px_timeline_color_kwargs = px_timeline_color_kwargs or {}
    
    if df.empty:
    
        raise NotImplementedError
    
    else:
        
        # Plotly can"t handle ns
        # plotly.express._core == 5.7.0, line 1681
        # args["data_frame"][args["x_end"]] = (x_end - x_start).astype("timedelta64[ms]")
        df.loc[:, COL_SUBSTAGE_DATE_START] = (
            df[COL_SUBSTAGE_DATE_START].dt.strftime(GANTT_XAXIS_DATETIME_FORMAT)
        )
    
        df.loc[:, COL_SUBSTAGE_DATE_END] = (
            df[COL_SUBSTAGE_DATE_END].dt.strftime(GANTT_XAXIS_DATETIME_FORMAT)
        )
    
        timeline_stages_fig = px.timeline(
            data_frame=df,
            x_start=COL_SUBSTAGE_DATE_START,
            x_end=COL_SUBSTAGE_DATE_END,
            y=col_y,
            text="y_labels",
            **px_timeline_color_kwargs,
        )
        
        _len = len(df)
        _start = pd.to_datetime(df[COL_SUBSTAGE_DATE_START], format=GANTT_XAXIS_DATETIME_FORMAT)
        _end = pd.to_datetime(df[COL_SUBSTAGE_DATE_END], format=GANTT_XAXIS_DATETIME_FORMAT)
        _duration = (_end - _start).astype("timedelta64[s]")
        
        _hovertext = (
            pd.Series(["Stage: <b>"] * _len)
            + df["y_labels"]
            + pd.Series(["</b>"] * _len)
            + pd.Series(["<br>"] * _len)
            + _start.dt.strftime("%H:%M:%S")
            + pd.Series([" to "] * _len)
            + _end.dt.strftime("%H:%M:%S")
            + pd.Series([" ("] * _len)
            + _duration.map("{:.0f}".format)
            + pd.Series([" sec)"] * _len)
        )
    
        timeline_stages_fig.update_traces(
            hovertext=_hovertext,
            hovertemplate=(
                "%{hovertext}"
                "<extra></extra>"
            ),
            marker_line=dict(
                width=1,
                color="white",
            ),
            marker_color="black",
            textposition="inside",
            textangle=0,
            insidetextanchor="middle",
        )
    
        timeline_stages_trace = list(
            timeline_stages_fig.select_traces()
        )
    
        assert len(timeline_stages_trace) == 1
        timeline_stages_trace = timeline_stages_trace[0]
        
        fig.add_trace(
            timeline_stages_trace,
            row=row,
            col=1,
        )
    
        fig.update_xaxes(type="date")



import logging
from typing import List, Tuple

import pandas as pd
from numpy import datetime64

from spark_sight.data_references import (
    COL_TASK_DATE_START,
    COL_TASK_DATE_END,
    COL_ID_STAGE,
    COL_STAGE_ASOFTASKS_DATE_START,
    COL_STAGE_ASOFTASKS_DATE_END,
    COL_SUBSTAGE_DATE_START,
    COL_SUBSTAGE_DATE_END,
    COL_SUBSTAGE_DATE_INTERVAL,
    COL_SUBSTAGE_DURATION,
    COL_SPLIT_BORDERS_LIST,
    COL_STAGE_DATE_START,
    COL_STAGE_DATE_END,
    COL_STAGE_DURATION,
)
from spark_sight.log_parse.main import extract_event_stage


def check_tasks_are_split_correctly(
    df: pd.DataFrame,
) -> None:
    not_split_correcly = (
        df[
            (
                df[COL_TASK_DATE_START] <= (
                df[COL_SUBSTAGE_DATE_INTERVAL].map(
                    lambda _substage_interval: _substage_interval.left
                )
            )
            )
            | (
                df[COL_TASK_DATE_END] > (
                df[COL_SUBSTAGE_DATE_INTERVAL].map(
                    lambda _substage_interval: _substage_interval.right
                )
            )
            )
        ]
    )
    
    if not not_split_correcly.empty:
        raise ValueError(
            f"Found tasks not split correctly: \n{not_split_correcly}"
        )


def aggregate_tasks_in_substages(
    df: pd.DataFrame,
    borders_all: List[datetime64],
    metrics: List[str],
    cols_groupby: List[str],
) -> pd.DataFrame:
    """Aggregate task metrics within the substage intervals.
    
    Sums the metrics in the substage intervals,
    and computes the efficiency based on the substage duration.
    
    A substage is the interval between two consecutive borders.
    
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing task information already split
        across the borders.
        The id task column may not be primary key.
    borders_all : list of datetimes
        List of borders, including external borders
        of first stage start and last stage end.
    metrics : list of str
        List of metric strings to aggregate (sum).
    cols_groupby : list of str
        Columns to group by to aggregate the metrics.

    Returns
    -------
    pd.DataFrame
        Aggregate information regarding tasks in substages.
        Contains:
        * substage interval
        * unique stage ids of the tasks that have been aggregated
        * aggregate metric
    
    """
    df.loc[:, COL_SUBSTAGE_DATE_INTERVAL] = (
        pd.cut(
            df[COL_TASK_DATE_END],
            borders_all,
        )
    )

    check_tasks_are_split_correctly(df)
    
    grouped = (
        df
        .groupby(
            cols_groupby,
        )
        .agg(
            {
                COL_ID_STAGE: "unique",
                **{
                    _metric: "sum"
                    for _metric in metrics
                },
            }
        )
        .reset_index()
    )

    grouped = grouped[
        grouped[COL_ID_STAGE].notna()
    ]
    
    grouped = (
        grouped[
            grouped[COL_ID_STAGE].map(len) > 0
        ]
        .copy().reset_index(drop=True)
    )
    
    grouped.loc[:, COL_SUBSTAGE_DATE_START] = (
        grouped.apply(
            lambda _row: _row[COL_SUBSTAGE_DATE_INTERVAL].left,
            axis=1,
        )
    )
    
    grouped.loc[:, COL_SUBSTAGE_DATE_END] = (
        grouped.apply(
            lambda _row: _row[COL_SUBSTAGE_DATE_INTERVAL].right,
            axis=1,
        )
    )
    
    grouped.loc[:, COL_SUBSTAGE_DURATION] = (
        grouped.apply(
            (
                lambda _row: (
                    (
                        _row[COL_SUBSTAGE_DATE_END]
                        - _row[COL_SUBSTAGE_DATE_START]
                    ).delta
                )
            ),
            axis=1,
        )
    )
    
    return grouped


def split_on_borders(
    task_info: pd.DataFrame,
    borders_all: List[datetime64],
    metrics: List[str],
) -> pd.DataFrame:
    """Split tasks running on the inner borders.
    
    Iteratively determines tasks that are running across an inner border
    of the borders input list, and replaces the dataframe row of the task
    with two rows where the task has been split in two.
    
    Parameters
    ----------
    task_info : pd.DataFrame
        Dataframe containing task information.
        The id task column is expected to be primary key.
    borders_all : list of datetimes
        List of borders, including external borders
        of first stage start and last stage end.
    metrics : list of str
        List of metric strings to split.

    Returns
    -------
    pd.DataFrame
        Dataframe where task rows of tasks on inner borders have been split.
        A new column containing the information of which borders
        have been used to apply the split is added.
        The id task column may not be primary key.
    
    """
    threshold_metrics_discard = 1e-2
    borders_inner = borders_all[1:-1]
    df_split = task_info.copy()
    
    df_split.loc[:, COL_SPLIT_BORDERS_LIST] = pd.Series(
        [
            [] for _ in range(len(df_split))
        ]
    )
    
    for index_border, border in enumerate(borders_inner):
        logging.debug(f"Border {border}")
        
        # Remap to index of borders input list
        index_border += 1
        
        on_the_border = (
            df_split[
                (df_split[COL_TASK_DATE_START] <= border)
                & (df_split[COL_TASK_DATE_END] > border)
            ]
        )
        
        line_just_split_index = []
        line_just_split_rows_split = []
        
        for on_the_border_row in on_the_border.itertuples():
            to_split_index = getattr(on_the_border_row, "Index")
            to_split_row = on_the_border.loc[to_split_index].copy()
            
            (
                to_split_row_left,
                to_split_row_right,
            ) = split_single_border(
                border=border,
                to_split_row=to_split_row,
                metrics=metrics
                )
            
            line_just_split_index.append(
                to_split_index
            )
            
            # Discard split resulting in all metrics < threshold
            if any(
                to_split_row_left[_metric] > float(threshold_metrics_discard)
                for _metric in metrics
            ):
                line_just_split_rows_split.append(to_split_row_left)
            else:
                logging.debug(
                    f"Discarding split\n{to_split_row_left}"
                )
                
            if any(
                to_split_row_right[_metric] > float(threshold_metrics_discard)
                for _metric in metrics
            ):
                line_just_split_rows_split.append(to_split_row_right)
            else:
                logging.debug(
                    f"Discarding split\n{to_split_row_right}"
                )
                
        df_split = df_split.drop(line_just_split_index)
        
        df_split = df_split.append(
            line_just_split_rows_split,
            ignore_index=True,
        )
        
        df_split = df_split.reset_index(drop=True)

    return df_split


def split_row_in_two(
    to_split_row_generic,
) -> (pd.Series, pd.Series):
    to_split_row_left = to_split_row_generic.copy()
    to_split_row_right = to_split_row_generic.copy()
    
    return (
        to_split_row_left,
        to_split_row_right,
    )


def split_single_border(
    border: datetime64,
    to_split_row: pd.Series,
    metrics: List[str],
) -> Tuple[pd.Series, pd.Series]:
    """Split the task across the border.
    
    Parameters
    ----------
    border : datetime64
        Border across which to split.
    to_split_row : pd.Series
        Task to split.
    metrics : list of str
        List of metric strings present in the task Pandas series
        to split proportionally to the new task lengths.
    
    Returns
    -------
    pd.Series
        New task, left split.
    pd.Series
        New task, right split.
    
    """

    border_left = None
    border_right = None
    
    if len(metrics) == 0:
        raise ValueError("Provided empty metrics")
    
    (
        _left,
        _right,
    ) = split_row_in_two(to_split_row)
    
    for metric in metrics:
        metric_total = _left[metric]
    
        border_left = border
        border_right = border + pd.Timedelta(1, unit="ns")
        
        reference_left = _left[COL_TASK_DATE_START]
        reference_right = _right[COL_TASK_DATE_END]
    
        duration_left = abs(reference_left - border_left)
        duration_right = abs(reference_right - border_right)
    
        duration_total_left = (
            _left[COL_TASK_DATE_END]
            - _left[COL_TASK_DATE_START]
        )

        duration_total_right = (
            _right[COL_TASK_DATE_END]
            - _right[COL_TASK_DATE_START]
        )
    
        _left[metric] = (
            metric_total
            * (
                float(duration_left.delta)
                / float(duration_total_left.delta)
            )
        )

        _right[metric] = (
            metric_total
            * (
                float(duration_right.delta)
                / float(duration_total_right.delta)
            )
        )
        
    _left[COL_TASK_DATE_END] = border_left
    _right[COL_TASK_DATE_START] = border_right
    
    return (
        _left,
        _right,
    )


def determine_borders_of_stages_asoftasks(
    task_info: pd.DataFrame,
) -> List[datetime64]:
    """Determine borders of stages as of tasks starting and ending.
    
    Determines
    * the start of a stage as the min among all tasks start dates
    * the end of a stage as the max among all tasks end dates
    
    In fact, logged stage timings are different from the associated task timings:
    
    * stage `Submission Time` <= first task `Launch Time`
    * stage `Completion Time` >= last task `Finish Time`
    
    Parameters
    ----------
    task_info : pd.DataFrame
        DataFrame of task information.

    Returns
    -------
    list
        List of borders, including external borders
        of stage start and stage end.
    
    """
    for id_stage in task_info[COL_ID_STAGE].unique():
        logging.debug(f"Stage {id_stage}")
        
        _index_stage = task_info[COL_ID_STAGE] == id_stage
        
        # Subtracting 1 ns because we consider task belonging to interval
        # iff > stage start
        stage_asoftasks_date_start = (
            task_info.loc[_index_stage][COL_TASK_DATE_START].agg("min")
            - pd.Timedelta(1, unit="ns")
        )
        
        # Subtracting 0 ns because we consider task belonging to interval
        # iff <= stage end
        stage_asoftasks_date_end = (
            task_info.loc[_index_stage][COL_TASK_DATE_END].agg("max")
        )
        
        logging.debug(
            f"Start {stage_asoftasks_date_start}"
            f", end {stage_asoftasks_date_end}"
        )
        
        task_info.loc[_index_stage, COL_STAGE_ASOFTASKS_DATE_START] = (
            stage_asoftasks_date_start
        )
        
        task_info.loc[_index_stage, COL_STAGE_ASOFTASKS_DATE_END] = (
            stage_asoftasks_date_end
        )
    
    borders_all = set(
        list(task_info[COL_STAGE_ASOFTASKS_DATE_START].unique())
        + list(task_info[COL_STAGE_ASOFTASKS_DATE_END].unique())
    )
    
    borders_all = sorted(list(borders_all))
    
    logging.debug(f"Borders {borders_all}")
    
    return borders_all


def create_duration_stage(
    lines_stages,
    stage_ids,
):
    data_stage_single = []
    
    for stage_id in stage_ids:
        stage = extract_event_stage(lines_stages, stage_id=stage_id)
        
        stage_duration = (
            (
                stage["Stage Info"]["Completion Time"]
                - stage["Stage Info"]["Submission Time"]
            )
            / 1e3
        )
        
        # TODOish Launch time of any task related to the stage
        # can be very far from stage submission time
        start_date = pd.to_datetime(
            stage["Stage Info"]["Submission Time"] * 1e6
        )
        end_date = pd.to_datetime(stage["Stage Info"]["Completion Time"] * 1e6)
        
        data_stage_single.append(
            {
                COL_ID_STAGE: stage_id,
                COL_STAGE_DATE_START: start_date,
                COL_STAGE_DATE_END: end_date,
                COL_STAGE_DURATION: stage_duration,
            }
        )
    
    _df_stage = pd.DataFrame(data_stage_single)
    
    return _df_stage



import logging
from json import loads as json_loads
from typing import Optional

from requests import get, codes, Response

from spark_sight import __version__ as spark_sight_version


try:
    from packaging.version import parse
except ImportError:
    from pip._vendor.packaging.version import parse


def configure_pandas(
    pandas,
    infinite_rows: bool = False,
    additional_params: bool = False,
    float_precision: int = 2,
):
    """Configure Pandas module.

    Args:
        pandas: Pandas module
            where to apply the configuration.
        infinite_rows (bool): Whether to configure Pandas
            to display all rows.
        additional_params (bool): Whether to configure Pandas
            to display wider and justified columns.

    Returns:

    """
    pandas.set_option("display.expand_frame_repr", False)
    pandas.options.display.width = 0
    pandas.set_option("display.max_columns", None)
    pandas.options.display.float_format = (
        "{:,." + str(float_precision) + "f}"
    ).format
    
    if infinite_rows:
        pandas.set_option("display.max_rows", None)
    
    if additional_params:
        pandas.set_option("display.max_colwidth", 100)
        pandas.set_option("display.column_space", 40)
        pandas.set_option("display.colheader_justify", "left")


def is_latest_version():
    response: Optional[Response]
    
    try:
        response = get(
            'https://pypi.python.org/pypi/spark-sight/json',
            timeout=1,
        )
    except ConnectionError:
        response = None
    except Exception as e:
        logging.debug(
            f"Getting latest version, not connection error: {e}"
        )
        response = None
    
    if response is None:
        return True

    if response.status_code == codes.ok:
        j = json_loads(response.text.encode(response.encoding))
        releases = j.get('releases', [])
        assert releases, "No releases found"
        return spark_sight_version == max(
            parse(release)
            for release in releases
            if not parse(release).is_prerelease
        )
